#define _GNU_SOURCE        /* or _BSD_SOURCE or _SVID_SOURCE */
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/time.h>

#define N 10000

void* otherthread(void* dummy)
{
    unsigned i;
    while (1) {
	sched_yield();
	pthread_testcancel();
    }
    return NULL;
}

void timesched(int fd)
{
    char buf[5];
    struct timeval tv0, tv1;
    double t;
    int i;
    pthread_t thr;

    pthread_create(&thr, NULL, otherthread, NULL);
    gettimeofday(&tv0, NULL);
    for (i=0;i<N;i++) {
	sched_yield();
    }
    gettimeofday(&tv1, NULL);
    pthread_cancel(thr);
    pthread_join(thr, NULL);
    t = tv1.tv_sec - tv0.tv_sec + (tv1.tv_usec- tv0.tv_usec)/1000000.0;
    printf("Yield Time: %fs, Latency: %fus\n",  t, 1000000.0*t/N);
}

static pthread_spinlock_t slock;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

#define LOCK do {\
    if (lock == 1) pthread_spin_lock(&slock);\
    else if (lock == 2) pthread_mutex_lock(&mutex);} while(0)

#define UNLOCK do {\
    if (lock == 1) pthread_spin_unlock(&slock);\
    else if (lock == 2) pthread_mutex_unlock(&mutex);} while(0)

static void *mapthread(void *dummy)
{
    int i, lock = *(int*)dummy;
    struct timeval tv0, tv1;
    double t;
    gettimeofday(&tv0, NULL);
    for (i=0;i<N;i++) {
	LOCK;
	char *h = mmap(NULL, 1024*4096, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
	UNLOCK;
	LOCK;
	*((volatile char*)h) = 0;
	UNLOCK;
	LOCK;
	munmap(h, 1024*4096);
	UNLOCK;
    }
    gettimeofday(&tv1, NULL);
    t = tv1.tv_sec - tv0.tv_sec + (tv1.tv_usec- tv0.tv_usec)/1000000.0;
    printf("Mmap Time [%s]: %fs, Latency: %fus\n",
	   lock == 1 ? "spinlock" : lock == 2 ? "mutex" : "nolock",
	   t, 1000000.0*t/N);
}

void timemmap(int lock)
{
    char buf[5];
    double t;
    int i;
    pthread_t thr1, thr2, thr3, thr4;

    pthread_create(&thr1, NULL, mapthread, &lock);
    pthread_create(&thr2, NULL, mapthread, &lock);
    pthread_create(&thr3, NULL, mapthread, &lock);
    pthread_create(&thr4, NULL, mapthread, &lock);
    mapthread(&lock);
    pthread_join(thr1, NULL);
    pthread_join(thr2, NULL);
    pthread_join(thr3, NULL);
    pthread_join(thr4, NULL);
}

int main()
{
    pthread_spin_init(&slock, PTHREAD_PROCESS_PRIVATE);
    timesched(0);
    timemmap(0);
    timemmap(1);
    timemmap(2);
    pthread_spin_destroy(&slock);
    return 0;
}
